;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "annalist" "20260531.1558"
  "Record and display information such as keybindings."
  '((emacs  "24.4")
    (cl-lib "0.5"))
  :url "https://github.com/noctuid/annalist.el"
  :commit "0d958732b710a8e9edc4c70b2318570e1c7d4923"
  :revdesc "0d958732b710"
  :keywords '("convenience" "tools" "keybindings" "org")
  :authors '(("Fox Kiester" . "noct@posteo.net"))
  :maintainers '(("Fox Kiester" . "noct@posteo.net")))
